@@ -1,34 +1,25 @@
-/* radare - LGPL - Copyright 2010-2020 - pancake, maijin */
-
-#include <r_types.h>
-#include <r_list.h>
-#include <r_flag.h>
-#include <r_core.h>
-#define USE_R2 1
-#include <spp/spp.h>
+/* radare - LGPL - Copyright 2010-2024 - pancake, rhl */
 
 #define PROJECT_EXPERIMENTAL 0
 
+// R2R db/cmd/projects
+
+#include <r_core.h>
+#include <rvc.h>
+// required to make spp use RStrBuf instead of SStrBuf
+#define USE_R2 1
+#include <spp/spp.h>
+
 // project apis to be used from cmd_project.c
+// TODO: Use .zrp as in zipped radare project
 
 static bool is_valid_project_name(const char *name) {
-	int i;
-	if (r_str_endswith (name, ".zip")) {
+	if (r_str_len_utf8 (name) >= 64) {
 		return false;
 	}
-	for (i = 0; name[i]; i++) {
-		switch (name[i]) {
-		case '\\': // for w32
-		case '.':
-		case '_':
-		case ':':
-		case '-':
-			continue;
-		}
-		if (isalpha ((unsigned char)name[i])) {
-			continue;
-		}
-		if (IS_DIGIT (name[i])) {
+	const char * const extention = r_str_endswith (name, ".zip")? r_str_last (name, ".zip"): NULL;
+	for (; *name && name != extention; name++) {
+		if (isdigit (*name) || islower (*name) || *name == '_') {
 			continue;
 		}
 		return false;
@@ -37,7 +28,10 @@ static bool is_valid_project_name(const char *name) {
 }
 
 static char *get_project_script_path(RCore *core, const char *file) {
-	r_return_val_if_fail (core && !R_STR_ISEMPTY (file), NULL);
+	R_RETURN_VAL_IF_FAIL (core && file, NULL);
+	if (!*file) {
+		return NULL;
+	}
 	const char *magic = "# r2 rdb project file";
 	char *data, *prjfile;
 	if (r_file_is_abspath (file)) {
@@ -67,7 +61,7 @@ static int make_projects_directory(RCore *core) {
 	char *prjdir = r_file_abspath (r_config_get (core->config, "dir.projects"));
 	int ret = r_sys_mkdirp (prjdir);
 	if (!ret) {
-		eprintf ("Cannot mkdir dir.projects\n");
+		R_LOG_ERROR ("Cannot mkdir dir.projects");
 	}
 	free (prjdir);
 	return ret;
@@ -75,7 +69,7 @@ static int make_projects_directory(RCore *core) {
 
 R_API bool r_core_is_project(RCore *core, const char *name) {
 	bool ret = false;
-	if (name && *name && *name != '.') {
+	if (R_STR_ISNOTEMPTY (name) && *name != '.') {
 		char *path = get_project_script_path (core, name);
 		if (!path) {
 			return false;
@@ -93,32 +87,32 @@ R_API bool r_core_is_project(RCore *core, const char *name) {
 	return ret;
 }
 
-R_API int r_core_project_cat(RCore *core, const char *name) {
+R_API void r_core_project_cat(RCore *core, const char *name) {
+	r_core_return_value (core, R_CMD_RC_FAILURE);
 	char *path = get_project_script_path (core, name);
 	if (path) {
 		char *data = r_file_slurp (path, NULL);
 		if (data) {
 			r_cons_println (data);
 			free (data);
+			r_core_return_value (core, R_CMD_RC_SUCCESS);
 		}
+		free (path);
 	}
-	free (path);
-	return 0;
 }
 
 R_API int r_core_project_list(RCore *core, int mode) {
 	PJ *pj = NULL;
 	RListIter *iter;
-	RList *list;
 
 	char *foo, *path = r_file_abspath (r_config_get (core->config, "dir.projects"));
 	if (!path) {
 		return 0;
 	}
-	list = r_sys_dir (path);
+	RList *list = r_sys_dir (path);
 	switch (mode) {
 	case 'j':
-		pj = pj_new ();
+		pj = r_core_pj_new (core);
 		if (!pj) {
 			break;
 		}
@@ -146,68 +140,30 @@ R_API int r_core_project_list(RCore *core, int mode) {
 	return 0;
 }
 
-static inline void remove_project_file(char * path) {
-	if (r_file_exists (path)) {
-		r_file_rm (path);
-		eprintf ("rm %s\n", path);
-	}
-}
-
-static inline void remove_notes_file(const char *pd) {
-	char *notes_txt = r_file_new (pd, "notes.txt", NULL);
-	if (r_file_exists (notes_txt)) {
-		r_file_rm (notes_txt);
-		eprintf ("rm %s\n", notes_txt);
-	}
-	free (notes_txt);
-}
-
-static inline void remove_rop_directory(const char *prj_dir) {
-	char *rop_d = r_file_new (prj_dir, "rop.d", NULL);
-
-	if (r_file_is_directory (rop_d)) {
-		char *f;
-		RListIter *iter;
-		RList *files = r_sys_dir (rop_d);
-		r_list_foreach (files, iter, f) {
-			char *filepath = r_str_append (strdup (rop_d), R_SYS_DIR);
-			filepath = r_str_append (filepath, f);
-			if (!r_file_is_directory (filepath)) {
-				eprintf ("rm %s\n", filepath);
-				r_file_rm (filepath);
-			}
-
-			free (filepath);
-		}
-		r_file_rm (rop_d);
-		eprintf ("rm %s\n", rop_d);
-		r_list_free (files);
-	}
-
-	free (rop_d);
-}
 R_API int r_core_project_delete(RCore *core, const char *prjfile) {
 	if (r_sandbox_enable (0)) {
-		eprintf ("Cannot delete project in sandbox mode\n");
+		R_LOG_ERROR ("Cannot delete project in sandbox mode");
 		return 0;
 	}
 	char *path = get_project_script_path (core, prjfile);
 	if (!path) {
-		eprintf ("Invalid project name '%s'\n", prjfile);
+		R_LOG_ERROR ("Invalid project name '%s'", prjfile);
 		return false;
 	}
 	if (r_core_is_project (core, prjfile)) {
 		char *prj_dir = r_file_dirname (path);
 		if (!prj_dir) {
-			eprintf ("Cannot resolve directory\n");
+			R_LOG_ERROR ("Cannot resolve directory");
 			free (path);
 			return false;
 		}
-		remove_project_file (path);
-		remove_notes_file (prj_dir);
-		remove_rop_directory (prj_dir);
-		// remove directory only if it's empty
-		r_file_rm (prj_dir);
+		bool must_rm = true;
+		if (r_config_get_b (core->config, "scr.interactive")) {
+			must_rm = r_cons_yesno ('y', "Confirm project deletion? (Y/n)");
+		}
+		if (must_rm) {
+			r_file_rm_rf (prj_dir);
+		}
 		free (prj_dir);
 	}
 	free (path);
@@ -215,7 +171,7 @@ R_API int r_core_project_delete(RCore *core, const char *prjfile) {
 }
 
 static bool load_project_rop(RCore *core, const char *prjfile) {
-	r_return_val_if_fail (core && !R_STR_ISEMPTY (prjfile), false);
+	R_RETURN_VAL_IF_FAIL (core && R_STR_ISNOTEMPTY (prjfile), false);
 	char *path, *db = NULL, *path_ns;
 	bool found = 0;
 	SdbListIter *it;
@@ -230,10 +186,8 @@ static bool load_project_rop(RCore *core, const char *prjfile) {
 
 	char *rc_path = get_project_script_path (core, prjfile);
 	char *prj_dir = r_file_dirname (rc_path);
-
+	R_FREE (rc_path);
 	if (r_str_endswith (prjfile, R_SYS_DIR "rc.r2")) {
-		// XXX
-		eprintf ("ENDS WITH\n");
 		path = strdup (prjfile);
 		path[strlen (path) - 3] = 0;
 	} else if (r_file_fexists ("%s%s%src.r2", R_SYS_DIR, prj_dir, prjfile)) {
@@ -243,7 +197,6 @@ static bool load_project_rop(RCore *core, const char *prjfile) {
 			db = r_str_newf ("%s.d", prjfile);
 			if (!db) {
 				free (prj_dir);
-				free (rc_path);
 				return false;
 			}
 			path = strdup (db);
@@ -251,7 +204,6 @@ static bool load_project_rop(RCore *core, const char *prjfile) {
 			db = r_str_newf ("%s" R_SYS_DIR "%s.d", prj_dir, prjfile);
 			if (!db) {
 				free (prj_dir);
-				free (rc_path);
 				return false;
 			}
 			path = r_file_abspath (db);
@@ -260,11 +212,10 @@ static bool load_project_rop(RCore *core, const char *prjfile) {
 	if (!path) {
 		free (db);
 		free (prj_dir);
-		free (rc_path);
 		return false;
 	}
 	if (rop_db) {
-		ls_foreach (core->sdb->ns, it, ns){
+		ls_foreach (core->sdb->ns, it, ns) {
 			if (ns->sdb == rop_db) {
 				ls_delete (core->sdb->ns, it);
 				found = true;
@@ -280,7 +231,6 @@ static bool load_project_rop(RCore *core, const char *prjfile) {
 		free (db);
 		free (path);
 		free (prj_dir);
-		free (rc_path);
 		return false;
 	}
 	sdb_ns_set (core->sdb, "rop", rop_db);
@@ -308,17 +258,14 @@ static bool load_project_rop(RCore *core, const char *prjfile) {
 	free (path_ns);
 	free (db);
 	free (prj_dir);
-	free (rc_path);
 	return true;
 }
 
 R_API void r_core_project_execute_cmds(RCore *core, const char *prjfile) {
 	char *str = r_core_project_notes_file (core, prjfile);
 	char *data = r_file_slurp (str, NULL);
-	if (!data) {
-		free (str);
-		return;
-	}
+	free (str);
+	R_RETURN_IF_FAIL (data);
 	Output out;
 	out.fout = NULL;
 	out.cout = r_strbuf_new (NULL);
@@ -328,49 +275,78 @@ R_API void r_core_project_execute_cmds(RCore *core, const char *prjfile) {
 	spp_eval (data, &out);
 	free (data);
 	data = strdup (r_strbuf_get (out.cout));
-	char *bol = strtok (data, "\n");
+	char *save_ptr = NULL;
+	char *bol = r_str_tok_r (data, "\n", &save_ptr);
 	while (bol) {
 		if (bol[0] == ':') {
 			r_core_cmd0 (core, bol + 1);
 		}
-		bol = strtok (NULL, "\n");
+		bol = r_str_tok_r (NULL, "\n", &save_ptr);
 	}
 	free (data);
-	free (str);
 }
 
 typedef struct {
 	RCore *core;
 	char *prj_name;
 	char *rc_path;
-} projectState;
+} ProjectState;
 
 static bool r_core_project_load(RCore *core, const char *prj_name, const char *rcpath) {
+	R_RETURN_VAL_IF_FAIL (core, false);
 	if (R_STR_ISEMPTY (prj_name)) {
 		prj_name = r_core_project_name (core, rcpath);
 	}
 	if (r_project_is_loaded (core->prj)) {
-		eprintf ("o--;e prj.name=\n");
-		return false;
+		R_LOG_INFO ("o--;e prj.name=");
+	//	return false;
 	}
 	if (!r_project_open (core->prj, prj_name, rcpath)) {
 		return false;
 	}
-	const bool cfg_fortunes = r_config_get_i (core->config, "cfg.fortunes");
+	const bool cfg_fortunes = r_config_get_b (core->config, "cfg.fortunes");
 	const bool scr_interactive = r_cons_is_interactive ();
-	const bool scr_prompt = r_config_get_i (core->config, "scr.prompt");
+	const bool scr_prompt = r_config_get_b (core->config, "scr.prompt");
 	(void) load_project_rop (core, prj_name);
-	bool ret = r_core_cmd_file (core, rcpath);
-	r_config_set_i (core->config, "cfg.fortunes", cfg_fortunes);
-	r_config_set_i (core->config, "scr.interactive", scr_interactive);
-	r_config_set_i (core->config, "scr.prompt", scr_prompt);
+	const bool sandy = r_config_get_b (core->config, "prj.sandbox");
+	bool ret = false;
+	if (sandy) {
+		// enable sandbox (only allow file access, no network or program exec)
+		// projects can also tweak the cmd. eval vars to run code after the project is loaded
+		// users must be careful on that.
+		int oldgrain = r_sandbox_grain (R_SANDBOX_GRAIN_DISK | R_SANDBOX_GRAIN_FILES);
+		r_sandbox_enable (true);
+		ret = r_core_cmd_file (core, rcpath);
+		r_sandbox_disable (true);
+		r_sandbox_grain (oldgrain);
+	} else {
+		ret = r_core_cmd_file (core, rcpath);
+	}
+	char *prj_path = r_file_dirname (rcpath);
+	if (prj_path) {
+		//check if the project uses git
+		Rvc *vc = rvc_open (prj_path, RVC_TYPE_GIT);
+		core->prj->rvc = vc;
+		free (prj_path);
+	} else {
+		R_LOG_ERROR ("Failed to load rvc");
+	}
+	if (r_config_get_b (core->config, "prj.history")) {
+		char *file = r_file_new (prj_path, "history", NULL);
+		r_line_hist_free (); // R2_600 - hist_reset ?
+		r_line_hist_load (file);
+		free (file);
+	}
+	r_config_set_b (core->config, "cfg.fortunes", cfg_fortunes);
+	r_config_set_b (core->config, "scr.interactive", scr_interactive);
+	r_config_set_b (core->config, "scr.prompt", scr_prompt);
 	r_config_bump (core->config, "asm.arch");
 	r_config_set (core->config, "prj.name", prj_name);
 	return ret;
 }
 
 static RThreadFunctionRet project_load_background(RThread *th) {
-	projectState *ps = th->user;
+	ProjectState *ps = th->user;
 	r_core_project_load (ps->core, ps->prj_name, ps->rc_path);
 	free (ps->prj_name);
 	free (ps->rc_path);
@@ -379,13 +355,13 @@ static RThreadFunctionRet project_load_background(RThread *th) {
 }
 
 R_API RThread *r_core_project_load_bg(RCore *core, const char *prj_name, const char *rc_path) {
-	projectState *ps = R_NEW0 (projectState);
+	ProjectState *ps = R_NEW0 (ProjectState);
 	ps->core = core;
 	ps->prj_name = strdup (prj_name);
 	ps->rc_path = strdup (rc_path);
 	RThread *th = r_th_new (project_load_background, ps, false);
 	if (th) {
-		r_th_start (th, true);
+		r_th_start (th);
 		char thname[32] = {0};
 		size_t thlen = R_MIN (strlen (prj_name), sizeof (thname) - 1);
 		r_str_ncpy (thname, prj_name, thlen);
@@ -395,51 +371,52 @@ R_API RThread *r_core_project_load_bg(RCore *core, const char *prj_name, const c
 }
 
 R_API bool r_core_project_open(RCore *core, const char *prj_path) {
-	r_return_val_if_fail (core && !R_STR_ISEMPTY (prj_path), false);
-	int ret, close_current_session = 1;
+	R_RETURN_VAL_IF_FAIL (core && !R_STR_ISEMPTY (prj_path), false);
+	bool interactive = r_config_get_b (core->config, "scr.interactive");
+	bool close_current_session = true;
+	bool ask_for_closing = true;
 	if (r_project_is_loaded (core->prj)) {
-		eprintf ("There's a project already opened\n");
-		bool ccs = r_cons_yesno ('y', "Close current session? (Y/n)");
-		if (ccs) {
-			r_core_cmd0 (core, "o--");
-		} else {
-			eprintf ("Project not loaded.\n");
+		R_LOG_ERROR ("There's a project already opened");
+		ask_for_closing = false;
+		bool ccs = interactive? r_cons_yesno ('y', "Close current session? (Y/n)"): true;
+		if (!ccs) {
+			R_LOG_ERROR ("Project not loaded");
 			return false;
 		}
+		r_core_cmd0 (core, "o--");
+		r_core_cmd0 (core, "P-");
 	}
 	char *prj_name = r_core_project_name (core, prj_path);
 	char *prj_script = get_project_script_path (core, prj_path);
 	if (!prj_script) {
-		eprintf ("Invalid project name '%s'\n", prj_path);
+		R_LOG_ERROR ("Invalid project name '%s'", prj_path);
 		return false;
 	}
-	if (r_project_is_loaded (core->prj)) {
+	if (ask_for_closing && r_project_is_loaded (core->prj)) {
 		if (r_cons_is_interactive ()) {
-			close_current_session = r_cons_yesno ('y', "Close current session? (Y/n)");
+			close_current_session = interactive
+				? r_cons_yesno ('y', "Close current session? (Y/n)")
+				: true;
 		}
 	}
 	if (close_current_session) {
-		r_core_cmd0 (core, "e prj.name=;o--");
+		r_config_set (core->config, "prj.name", "");
+		r_core_cmd0 (core, "o--");
 	}
 	/* load sdb stuff in here */
-	ret = r_core_project_load (core, prj_name, prj_script);
+	bool ret = r_core_project_load (core, prj_name, prj_script);
 	free (prj_name);
 	free (prj_script);
+	if (ret) {
+		r_core_project_undirty (core);
+	}
 	return ret;
 }
 
-R_API char *r_core_project_name(RCore *core, const char *prjfile) {
-	FILE *fd;
-	char buf[256], *file = NULL;
-	if (*prjfile != '/') {
-		return strdup (prjfile);
-	}
-	char *prj = get_project_script_path (core, prjfile);
-	if (!prj) {
-		eprintf ("Invalid project name '%s'\n", prjfile);
-		return NULL;
-	}
-	fd = r_sandbox_fopen (prj, "r");
+static char *get_project_name(const char *prj_script) {
+	char buf[1024];
+	char *file = NULL;
+	FILE *fd = r_sandbox_fopen (prj_script, "r");
 	if (fd) {
 		for (;;) {
 			if (!fgets (buf, sizeof (buf), fd)) {
@@ -448,24 +425,46 @@ R_API char *r_core_project_name(RCore *core, const char *prjfile) {
 			if (feof (fd)) {
 				break;
 			}
-			if (!strncmp (buf, "\"e prj.name = ", 14)) {
-				buf[strlen (buf) - 2] = 0;
-				file = r_str_new (buf + 14);
+			if (r_str_startswith (buf, "\"\"e prj.name = ")) {
+				file = strdup (buf + strlen ("\"\"e prj.name"));
+				break;
+			}
+			if (r_str_startswith (buf, "\"e prj.name = ")) {
+				// if (!strncmp (buf, "\"e prj.name = ", 14))
+				buf[strlen (buf) - 2] = 0; // remove trailing '"'
+				file = strdup (buf + 14);
+				break;
+			}
+			if (r_str_startswith (buf, "'e prj.name = ")) {
+				file = strdup (buf + strlen ("'e prj.name"));
 				break;
 			}
 		}
 		fclose (fd);
 	} else {
-		eprintf ("Cannot open project info (%s)\n", prj);
+		R_LOG_ERROR ("Cannot open project info (%s)", prj_script);
 	}
+	return file;
+}
+
+R_API char *r_core_project_name(RCore *core, const char *prjfile) {
+	if (*prjfile != '/') {
+		return strdup (prjfile);
+	}
+	char *prj = get_project_script_path (core, prjfile);
+	if (!prj) {
+		R_LOG_ERROR ("Invalid project name '%s'", prjfile);
+		return NULL;
+	}
+	char *file = get_project_name (prj);
 	free (prj);
 	if (R_STR_ISEMPTY (file)) {
 		free (file);
 		file = strdup (prjfile);
-		char *slash = (char *)r_str_lchr (file, R_SYS_DIR[0]); 
+		char *slash = (char *)r_str_lchr (file, R_SYS_DIR[0]);
 		if (slash) {
 			*slash = 0;
-			slash = (char *)r_str_lchr (file, R_SYS_DIR[0]); 
+			slash = (char *)r_str_lchr (file, R_SYS_DIR[0]);
 			if (slash) {
 				char *res = strdup (slash + 1);
 				free (file);
@@ -481,7 +480,7 @@ R_API char *r_core_project_name(RCore *core, const char *prjfile) {
 }
 
 #if PROJECT_EXPERIMENTAL
-static int fdc; // TODO: move into a struct passed to the foreach instead of global
+static R_TH_LOCAL int fdc; // TODO: move into a struct passed to the foreach instead of global
 
 static bool store_files_and_maps(RCore *core, RIODesc *desc, ut32 id) {
 	RList *maps = NULL;
@@ -489,13 +488,13 @@ static bool store_files_and_maps(RCore *core, RIODesc *desc, ut32 id) {
 	RIOMap *map;
 	if (desc) {
 		// reload bin info
-		r_cons_printf ("\"obf %s\"\n", desc->uri);
-		r_cons_printf ("\"ofs \\\"%s\\\" %s\"\n", desc->uri, r_str_rwx_i (desc->perm));
-		if ((maps = r_io_map_get_for_fd (core->io, id))) {
+		r_cons_printf ("'obf %s\n", desc->uri);
+		r_cons_printf ("'of \\\"%s\\\" %s\n", desc->uri, r_str_rwx_i (desc->perm));
+		if ((maps = r_io_map_get_by_fd (core->io, id))) { //wtf
 			r_list_foreach (maps, iter, map) {
-				r_cons_printf ("om %d 0x%"PFMT64x" 0x%"PFMT64x" 0x%"PFMT64x" %s%s%s\n", fdc,
-					r_io_map_begin (map), r_io_map_size (map), map->delta, r_str_rwx_i (map->perm),
-					map->name ? " " : "", r_str_get (map->name));
+				r_cons_printf ("om %d 0x%" PFMT64x " 0x%" PFMT64x " 0x%" PFMT64x " %s%s%s\n", fdc,
+						r_io_map_begin (map), r_io_map_size (map), map->delta, r_str_rwx_i (map->perm),
+						map->name? " " : "", r_str_get (map->name));
 			}
 			r_list_free (maps);
 		}
@@ -505,158 +504,193 @@ static bool store_files_and_maps(RCore *core, RIODesc *desc, ut32 id) {
 }
 #endif
 
-static bool project_save_script(RCore *core, const char *file, int opts) {
-	char *filename, *hl, *ohl = NULL;
-	int fdold;
+static void flush(RStrBuf *sb) {
+	char * s = r_cons_drain ();
+	if (s) {
+		r_strbuf_append (sb, s);
+		free (s);
+	}
+}
+
+R_API bool r_core_project_save_script(RCore *core, const char *file, int opts) {
+	char *hl, *ohl = NULL;
 
 	if (R_STR_ISEMPTY (file)) {
 		return false;
 	}
 
-	filename = r_str_word_get_first (file);
-	int fd = r_sandbox_open (file, O_BINARY | O_RDWR | O_CREAT | O_TRUNC, 0644);
-	if (fd == -1) {
-		free (filename);
-		return false;
-	}
+	char *filename = r_str_word_get_first (file);
 
 	hl = r_cons_singleton ()->highlight;
 	if (hl) {
 		ohl = strdup (hl);
 		r_cons_highlight (NULL);
 	}
-	fdold = r_cons_singleton ()->fdout;
-	r_cons_singleton ()->fdout = fd;
+	RStrBuf *sb = r_strbuf_new ("");
 	r_cons_singleton ()->context->is_interactive = false;
-	r_str_write (fd, "# r2 rdb project file\n");
+	r_cons_printf ("# r2 rdb project file\n");
 	// new behaviour to project load routine (see io maps below).
 	if (opts & R_CORE_PRJ_EVAL) {
-		r_str_write (fd, "# eval\n");
-		r_config_list (core->config, NULL, true);
-		r_cons_flush ();
+		r_cons_printf ("# eval\n");
+		r_config_list (core->config, NULL, 'r');
+		flush (sb);
 	}
 	r_core_cmd (core, "o*", 0);
+	r_core_cmd (core, "om*", 0);
 	r_core_cmd0 (core, "tcc*");
 	if (opts & R_CORE_PRJ_FCNS) {
-		r_str_write (fd, "# functions\n");
-		r_str_write (fd, "fs functions\n");
+		r_cons_printf ("# functions\n");
+		r_cons_printf ("fs functions\n");
 		r_core_cmd (core, "afl*", 0);
-		r_cons_flush ();
+		flush (sb);
+	}
+	{
+		r_cons_printf ("# registers\n");
+		r_core_cmd (core, "ar*", 0);
+		flush (sb);
+		r_core_cmd (core, "arR", 0);
+		flush (sb);
 	}
-
 	if (opts & R_CORE_PRJ_FLAGS) {
-		r_str_write (fd, "# flags\n");
+		r_cons_printf ("# flags\n");
 		r_flag_space_push (core->flags, NULL);
 		r_flag_list (core->flags, true, NULL);
 		r_flag_space_pop (core->flags);
-		r_cons_flush ();
+		flush (sb);
 	}
 #if PROJECT_EXPERIMENTAL
 	if (opts & R_CORE_PRJ_IO_MAPS && core->io && core->io->files) {
 		fdc = 3;
 		r_id_storage_foreach (core->io->files, (RIDStorageForeachCb)store_files_and_maps, core);
-		r_cons_flush ();
+		flush (sb);
 	}
 #endif
 	{
 		r_core_cmd (core, "fz*", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_META) {
-		r_str_write (fd, "# meta\n");
+		r_cons_printf ("# meta\n");
 		r_meta_print_list_all (core->anal, R_META_TYPE_ANY, 1, NULL);
-		r_cons_flush ();
+		flush (sb);
 		r_core_cmd (core, "fV*", 0);
-		r_cons_flush ();
+		flush (sb);
+		r_core_cmd (core, "ano*@@@F", 0);
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_XREFS) {
 		r_core_cmd (core, "ax*", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_FLAGS) {
 		r_core_cmd (core, "f.**", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_DBG_BREAK) {
 		r_core_cmd (core, "db*", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_ANAL_HINTS) {
 		r_core_cmd (core, "ah*", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_ANAL_TYPES) {
-		r_str_write (fd, "# types\n");
+		r_cons_printf ("# types\n");
 		r_core_cmd (core, "t*", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	if (opts & R_CORE_PRJ_ANAL_MACROS) {
-		r_str_write (fd, "# macros\n");
+		r_cons_printf ("# macros\n");
 		r_core_cmd (core, "(*", 0);
-		r_str_write (fd, "# aliases\n");
+		r_cons_printf ("# aliases\n");
 		r_core_cmd (core, "$*", 0);
-		r_cons_flush ();
+		flush (sb);
 	}
 	r_core_cmd (core, "wc*", 0);
 	if (opts & R_CORE_PRJ_ANAL_SEEK) {
-		r_cons_printf ("# seek\n"
-			"s 0x%08"PFMT64x "\n", core->offset);
-		r_cons_flush ();
+		r_cons_printf ("# seek\n" "s 0x%08" PFMT64x "\n", core->offset);
+		flush (sb);
 	}
-
-	r_cons_singleton ()->fdout = fdold;
 	r_cons_singleton ()->context->is_interactive = true;
+	flush (sb);
+	char *s = r_strbuf_drain (sb);
+	if (!strcmp (filename, "/dev/stdout")) {
+		r_cons_printf ("%s\n", s);
+	} else {
+		if (!r_file_dump (filename, (const ut8*)s, strlen (s), 0)) {
+			R_LOG_ERROR ("Cannot save file");
+		}
+	}
+	free (s);
 
 	if (ohl) {
 		r_cons_highlight (ohl);
 		free (ohl);
 	}
-
-	close (fd);
 	free (filename);
 
 	return true;
 }
 
-R_API bool r_core_project_save_script(RCore *core, const char *file, int opts) {
-	return project_save_script (core, file, opts);
+static void r_core_project_zip(RCore *core, const char *prj_dir) {
+	char *cwd = r_sys_getdir ();
+	const char *prj_name = r_file_basename (prj_dir);
+	if (r_sys_chdir (prj_dir)) {
+		if (!strchr (prj_name, '\'')) {
+			r_sys_chdir ("..");
+			char *zipfile = r_str_newf ("%s.zip", prj_name);
+			r_file_rm (zipfile);
+			// XXX use the ZIP api instead!
+			r_sys_cmdf ("zip -r %s %s", zipfile, prj_name);
+			free (zipfile);
+		} else {
+			R_LOG_WARN ("Command injection attempt?");
+		}
+	} else {
+		R_LOG_ERROR ("Cannot chdir %s", prj_dir);
+	}
+	r_sys_chdir (cwd);
+	free (cwd);
 }
 
 R_API bool r_core_project_save(RCore *core, const char *prj_name) {
+	R_RETURN_VAL_IF_FAIL (R_STR_ISNOTEMPTY (prj_name), false);
 	bool scr_null = false;
 	bool ret = true;
 	SdbListIter *it;
 	SdbNs *ns;
-	r_return_val_if_fail (prj_name && *prj_name, false);
+
+	if (r_config_get_b (core->config, "cfg.debug")) {
+		R_LOG_ERROR ("radare2 does not support projects on debugged bins");
+		return false;
+	}
 	char *script_path = get_project_script_path (core, prj_name);
-	if (r_config_get_i (core->config, "cfg.debug")) {
-		eprintf ("radare2 does not support projects on debugged bins.\n");
-		return false;
-	}
 	if (!script_path) {
-		eprintf ("Invalid project name '%s'\n", prj_name);
+		R_LOG_ERROR ("Invalid project name '%s'", prj_name);
 		return false;
 	}
-	char *prj_dir = NULL;
-	if (r_str_endswith (script_path, R_SYS_DIR "rc.r2")) {
-		/* new project format */
-		prj_dir = r_file_dirname (script_path);
-	} else {
-		prj_dir = r_str_newf ("%s.d", script_path);
-	}
+	char *prj_dir = r_str_endswith (script_path, R_SYS_DIR "rc.r2")
+		? r_file_dirname (script_path)
+		: r_str_newf ("%s.d", script_path);
 	if (r_file_exists (script_path)) {
 		if (r_file_is_directory (script_path)) {
-			eprintf ("Structural error: rc.r2 shouldnt be a directory.\n");
+			R_LOG_ERROR ("Structural error: rc.r2 shouldnt be a directory");
 		}
 	}
 	if (!prj_dir) {
 		prj_dir = strdup (prj_name);
 	}
-	if (!r_file_exists (prj_dir)) {
+	if (r_core_is_project (core, prj_name) && strcmp (prj_name, r_config_get (core->config, "prj.name"))) {
+		R_LOG_ERROR ("A project with this name already exists. Use P-%s to delete it", prj_name);
+		free (script_path);
+		free (prj_dir);
+		return false;
+	}
+	if (!r_file_is_directory (prj_dir)) {
 		r_sys_mkdirp (prj_dir);
 	}
-	if (r_config_get_i (core->config, "scr.null")) {
-		r_config_set_i (core->config, "scr.null", false);
+	if (r_config_get_b (core->config, "scr.null")) {
+		r_config_set_b (core->config, "scr.null", false);
 		scr_null = true;
 	}
 	make_projects_directory (core);
@@ -672,71 +706,85 @@ R_API bool r_core_project_save(RCore *core, const char *prj_name) {
 		}
 	}
 
-	if (!project_save_script (core, script_path, R_CORE_PRJ_ALL)) {
-		eprintf ("Cannot open '%s' for writing\n", prj_name);
+	r_config_set (core->config, "prj.name", prj_name);
+	if (!r_core_project_save_script (core, script_path, R_CORE_PRJ_ALL)) {
+		R_LOG_ERROR ("Cannot open '%s' for writing", prj_name);
 		ret = false;
+		r_config_set (core->config, "prj.name", "");
 	}
 
-	if (r_config_get_i (core->config, "prj.files")) {
-		eprintf ("TODO: prj.files: support copying more than one file into the project directory\n");
+	if (r_config_get_b (core->config, "prj.files")) {
 		char *bin_file = r_core_project_name (core, prj_name);
 		const char *bin_filename = r_file_basename (bin_file);
 		char *prj_bin_dir = r_str_newf ("%s" R_SYS_DIR "bin", prj_dir);
 		char *prj_bin_file = r_str_newf ("%s" R_SYS_DIR "%s", prj_bin_dir, bin_filename);
 		r_sys_mkdirp (prj_bin_dir);
 		if (!r_file_copy (bin_file, prj_bin_file)) {
-			eprintf ("Warning: Cannot copy '%s' into '%s'\n", bin_file, prj_bin_file);
+			R_LOG_WARN ("Cannot copy '%s' into '%s'", bin_file, prj_bin_file);
 		}
 		free (prj_bin_file);
 		free (prj_bin_dir);
 		free (bin_file);
 	}
-	if (r_config_get_i (core->config, "prj.git")) {
-		char *cwd = r_sys_getdir ();
-		char *git_dir = r_str_newf ("%s" R_SYS_DIR ".git", prj_dir);
-		if (r_sys_chdir (prj_dir)) {
-			if (!r_file_is_directory (git_dir)) {
-				r_sys_cmd ("git init");
+	if (core->prj->rvc || r_config_get_b (core->config, "prj.vc")) {
+		// assume that if the repo is not loaded, the repo doesn't exist
+		if (!core->prj->rvc) {
+			core->prj->rvc = rvc_open (prj_dir, RVC_TYPE_GIT);
+			if (!core->prj->rvc) {
+				R_LOG_WARN ("Cannot initialize git repositorty");
+				free (prj_dir);
+				free (script_path);
+				return false;
 			}
-			if (r_cons_is_interactive ()) {
-				r_sys_cmd ("git add * ; git commit -a");
+		}
+		RList *paths = r_list_new ();
+		if (paths) {
+			if (r_list_append (paths, prj_dir)) {
+				const char *author = r_config_get (core->config, "cfg.user");
+				const char *message = r_config_get (core->config, "prj.vc.message");
+				if (!rvc_commit (core->prj->rvc, message, author, paths)) {
+					r_list_free (paths);
+					free (prj_dir);
+					free (script_path);
+					return false;
+				}
+				rvc_save (core->prj->rvc);
 			} else {
-				r_sys_cmd ("git add * ; git commit -a -m commit");
+				r_list_free (paths);
+				free (prj_dir);
+				free (script_path);
+				return false;
 			}
 		} else {
-			eprintf ("Cannot chdir %s\n", prj_dir);
+			free (prj_dir);
+			free (script_path);
+			return false;
 		}
-		r_sys_chdir (cwd);
-		free (git_dir);
-		free (cwd);
 	}
-	if (r_config_get_i (core->config, "prj.zip")) {
-		char *cwd = r_sys_getdir ();
-		const char *prj_name = r_file_basename (prj_dir);
-		if (r_sys_chdir (prj_dir)) {
-			if (!strchr (prj_name, '\'')) {
-				r_sys_chdir ("..");
-				r_sys_cmdf ("rm -f '%s.zip'; zip -r '%s'.zip '%s'",
-					prj_name, prj_name, prj_name);
-			} else {
-				eprintf ("Command injection attempt?\n");
-			}
-		} else {
-			eprintf ("Cannot chdir %s\n", prj_dir);
-		}
-		r_sys_chdir (cwd);
-		free (cwd);
+	if (r_config_get_b (core->config, "prj.history")) {
+		char *history = r_core_cmd_str (core, "!!");
+		char *file = r_file_new (prj_dir, "history", NULL);
+		r_file_dump (file, (const ut8*)history, -1, false);
+		free (file);
+		free (history);
+	}
+	if (r_config_get_b (core->config, "prj.zip")) {
+		r_core_project_zip (core, prj_dir);
 	}
 	// LEAK : not always in heap free (prj_name);
-	free (prj_dir);
+	free (core->prj->path);
+	core->prj->path = prj_dir;
 	if (scr_null) {
-		r_config_set_i (core->config, "scr.null", true);
+		r_config_set_b (core->config, "scr.null", true);
 	}
 	free (script_path);
 	r_config_set (core->config, "prj.name", prj_name);
+	r_core_project_undirty (core);
 	return ret;
 }
 
+// dirty bits
+
 R_API char *r_core_project_notes_file(RCore *core, const char *prj_name) {
 	const char *prjdir = r_config_get (core->config, "dir.projects");
 	char *prjpath = r_file_abspath (prjdir);
@@ -744,3 +792,15 @@ R_API char *r_core_project_notes_file(RCore *core, const char *prj_name) {
 	free (prjpath);
 	return notes_txt;
 }
+
+R_API bool r_core_project_is_dirty(RCore *core) {
+	return !R_IS_DIRTY (core->config) && !R_IS_DIRTY (core->anal) && !R_IS_DIRTY (core->flags);
+}
+
+R_API void r_core_project_undirty(RCore *core) {
+	R_CRITICAL_ENTER (core);
+	core->config->is_dirty = false;
+	core->anal->is_dirty = false;
+	core->flags->is_dirty = false;
+	R_CRITICAL_LEAVE (core);
+}
