@@ -1,10 +1,9 @@
-/* radare2 - LGPL - Copyright 2011-2019 - pancake */
+/* radare2 - LGPL - Copyright 2011-2024 - pancake */
+
+#define R_LOG_ORIGIN "fs"
 
 #include <r_fs.h>
-#include "config.h"
-#include "types.h"
-#include <errno.h>
-#include "../../shlr/grub/include/grub/msdos_partition.h"
+#include <config.h>
 
 #if WITH_GPL
 # ifndef USE_GRUB
@@ -12,13 +11,36 @@
 # endif
 #endif
 
+#if WITH_GPL && USE_GRUB
+#include "../../shlr/grub/include/grubfs.h"
+#include "../../shlr/grub/include/grub/msdos_partition.h"
+#endif
+
 R_LIB_VERSION (r_fs);
 
-static RFSPlugin* fs_static_plugins[] = {
+static const RFSPlugin* fs_static_plugins[] = {
 	R_FS_STATIC_PLUGINS
 };
 
-R_API RFS* r_fs_new(void) {
+static const RFSType fstypes[] = {
+	{ "hfs", 0x400, "BD", 2, 0, 0, 0x400 },
+	{ "hfsplus", 0x400, "H+", 2, 0, 0, 0x400 },
+	{ "fat", 0x36, "FAT12", 5, 0, 0, 0 },
+	{ "fat", 0x52, "FAT32", 5, 0, 0, 0 },
+	{ "ext2", 0x438, "\x53\xef", 2, 0, 0, 0 },
+	{ "btrfs", 0x10040, "_BHRfS_M", 8, 0, 0, 0x0 },
+	{ "iso9660", 0x8000, "\x01" "CD0", 4, 0, 0, 0x8000 },
+	{ NULL }
+};
+
+R_API R_MUSTUSE const RFSType* r_fs_type_index(int i) {
+	if (i < 0 || i >= R_ARRAY_SIZE (fstypes)) {
+		return NULL;
+	}
+	return &fstypes[i];
+}
+
+R_API R_MUSTUSE RFS* r_fs_new(void) {
 	int i;
 	RFSPlugin* static_plugin;
 	RFS* fs = R_NEW0 (RFS);
@@ -43,7 +65,7 @@ R_API RFS* r_fs_new(void) {
 				continue;
 			}
 			memcpy (static_plugin, fs_static_plugins[i], sizeof (RFSPlugin));
-			r_fs_add (fs, static_plugin);
+			r_fs_plugin_add (fs, static_plugin);
 			free (static_plugin);
 		}
 	}
@@ -51,13 +73,11 @@ R_API RFS* r_fs_new(void) {
 }
 
 R_API RFSPlugin* r_fs_plugin_get(RFS* fs, const char* name) {
+	R_RETURN_VAL_IF_FAIL (fs && name, NULL);
 	RListIter* iter;
 	RFSPlugin* p;
-	if (!fs || !name) {
-		return NULL;
-	}
 	r_list_foreach (fs->plugins, iter, p) {
-		if (!strcmp (p->name, name)) {
+		if (!strcmp (p->meta.name, name)) {
 			return p;
 		}
 	}
@@ -65,29 +85,33 @@ R_API RFSPlugin* r_fs_plugin_get(RFS* fs, const char* name) {
 }
 
 R_API void r_fs_free(RFS* fs) {
-	if (!fs) {
-		return;
+	if (fs) {
+		//r_io_free (fs->iob.io);
+		//root makes use of plugin so revert to avoid UaF
+		r_list_free (fs->roots);
+		r_list_free (fs->plugins);
+		free (fs);
 	}
-	//r_io_free (fs->iob.io);
-	//root makes use of plugin so revert to avoid UaF
-	r_list_free (fs->roots);
-	r_list_free (fs->plugins);
-	free (fs);
 }
 
 /* plugins */
-R_API void r_fs_add(RFS* fs, RFSPlugin* p) {
-	// TODO: find coliding plugin name
-	if (p && p->init) {
+R_API bool r_fs_plugin_add(RFS* fs, RFSPlugin* p) {
+	R_RETURN_VAL_IF_FAIL (fs && p, false);
+	if (p->init) {
 		p->init ();
 	}
 	RFSPlugin* sp = R_NEW0 (RFSPlugin);
 	if (sp) {
-		if (p) {
-			memcpy (sp, p, sizeof (RFSPlugin));
-		}
+		memcpy (sp, p, sizeof (RFSPlugin));
 		r_list_append (fs->plugins, sp);
+		return true;
 	}
+	return false;
+}
+
+R_API bool r_fs_plugin_remove(RFS *fs, RFSPlugin *p) {
+	// XXX TODO
+	return true;
 }
 
 R_API void r_fs_del(RFS* fs, RFSPlugin* p) {
@@ -96,6 +120,7 @@ R_API void r_fs_del(RFS* fs, RFSPlugin* p) {
 
 /* mountpoint */
 R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 delta) {
+	R_RETURN_VAL_IF_FAIL (fs && fstype && path, NULL);
 	RFSPlugin* p;
 	RFSRoot* root;
 	RListIter* iter;
@@ -104,7 +129,7 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 	char *heapFsType = NULL;
 
 	if (path[0] != '/') {
-		eprintf ("r_fs_mount: invalid mountpoint %s\n", path);
+		R_LOG_ERROR ("invalid mountpoint %s", path);
 		return NULL;
 	}
 	if (!fstype || !*fstype) {
@@ -112,7 +137,7 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 		fstype = (const char *)heapFsType;
 	}
 	if (!(p = r_fs_plugin_get (fs, fstype))) {
-		// eprintf ("r_fs_mount: Invalid filesystem type\n");
+		R_LOG_ERROR ("Invalid filesystem type");
 		free (heapFsType);
 		return NULL;
 	}
@@ -123,7 +148,7 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 	}
 	r_str_trim_path (str);
 	if (*str && strchr (str + 1, '/')) {
-		eprintf ("r_fs_mount: mountpoint must have no subdirectories\n");
+		R_LOG_ERROR ("mountpoint must have no subdirectories");
 		free (heapFsType);
 		return NULL;
 	}
@@ -138,7 +163,7 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 			if (len > lenstr && root->path[lenstr] == '/') {
 				continue;
 			}
-			eprintf ("r_fs_mount: Invalid mount point\n");
+			R_LOG_ERROR ("Invalid mount point");
 			free (str);
 			free (heapFsType);
 			return NULL;
@@ -147,7 +172,7 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 	RFSFile* file = r_fs_open (fs, str, false);
 	if (file) {
 		r_fs_close (fs, file);
-		eprintf ("r_fs_mount: Invalid mount point\n");
+		R_LOG_ERROR ("Invalid mount point");
 		free (heapFsType);
 		free (str);
 		return NULL;
@@ -155,7 +180,7 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 	RList *list = r_fs_dir (fs, str);
 	if (!r_list_empty (list)) {
 		//XXX: list need free ??
-		eprintf ("r_fs_mount: Invalid mount point\n");
+		R_LOG_ERROR ("r_fs_mount: Invalid mount point");
 		free (str);
 		free (heapFsType);
 		return NULL;
@@ -165,14 +190,14 @@ R_API RFSRoot* r_fs_mount(RFS* fs, const char* fstype, const char* path, ut64 de
 	root->p = p;
 	root->iob = fs->iob;
 	root->cob = fs->cob;
-	if (!p->mount (root)) {
+	if (p->mount && !p->mount (root)) {
 		free (str);
 		free (heapFsType);
 		r_fs_root_free (root);
 		return NULL;
 	}
 	r_list_append (fs->roots, root);
-	eprintf ("Mounted %s on %s at 0x%" PFMT64x "\n", fstype, str, delta);
+	R_LOG_INFO ("Mounted %s on %s at 0x%" PFMT64x, fstype, str, delta);
 	free (str);
 	free (heapFsType);
 	return root;
@@ -183,16 +208,12 @@ static inline bool r_fs_match(const char* root, const char* path, int len) {
 }
 
 R_API bool r_fs_umount(RFS* fs, const char* path) {
-	int len;
+	R_RETURN_VAL_IF_FAIL (fs && path, false);
 	RFSRoot* root;
 	RListIter* iter, * riter = NULL;
 
-	if (!path) {
-		return false;
-	}
-
 	r_list_foreach (fs->roots, iter, root) {
-		len = strlen (root->path);
+		int len = strlen (root->path);
 		if (r_fs_match (path, root->path, len)) {
 			riter = iter;
 		}
@@ -205,7 +226,7 @@ R_API bool r_fs_umount(RFS* fs, const char* path) {
 }
 
 R_API RList* r_fs_root(RFS* fs, const char* p) {
-	RList* roots;
+	R_RETURN_VAL_IF_FAIL (fs && p, NULL);
 	RFSRoot* root;
 	RListIter* iter;
 	int len, olen;
@@ -213,7 +234,7 @@ R_API RList* r_fs_root(RFS* fs, const char* p) {
 	if (!path) {
 		return NULL;
 	}
-	roots = r_list_new ();
+	RList *roots = r_list_new ();
 	r_str_trim_path (path);
 	r_list_foreach (fs->roots, iter, root) {
 		len = strlen (root->path);
@@ -232,11 +253,13 @@ R_API RList* r_fs_root(RFS* fs, const char* p) {
 
 /* filez */
 R_API RFSFile* r_fs_open(RFS* fs, const char* p, bool create) {
+	R_RETURN_VAL_IF_FAIL (fs && p, NULL);
 	RFSRoot* root;
 	RListIter* iter;
 	RFSFile* f = NULL;
 	const char* dir;
 	char* path = r_str_trim_dup (p);
+	r_str_trim_path (path);
 	RList *roots = r_fs_root (fs, path);
 	if (!r_list_empty (roots)) {
 		r_list_foreach (roots, iter, root) {
@@ -267,62 +290,50 @@ R_API RFSFile* r_fs_open(RFS* fs, const char* p, bool create) {
 
 // NOTE: close doesnt free
 R_API void r_fs_close(RFS* fs, RFSFile* file) {
-	if (fs && file) {
-		R_FREE (file->data);
-		if (file->p && file->p->close) {
-			file->p->close (file);
-		}
+	R_RETURN_IF_FAIL (fs && file);
+	R_FREE (file->data);
+	if (file->p && file->p->close) {
+		file->p->close (file);
 	}
 }
 
 R_API int r_fs_write(RFS* fs, RFSFile* file, ut64 addr, const ut8 *data, int len) {
-	if (len < 1) {
-		return -1;
-	}
+	R_RETURN_VAL_IF_FAIL (fs && file && data && len >= 0, -1);
 	if (fs && file) {
-		// TODO: fill file->data ? looks like dupe of rbuffer 
+		// TODO: fill file->data ? looks like dupe of rbuffer
 		if (file->p && file->p->write) {
-			return file->p->write (file, addr, data, len);;
+			return file->p->write (file, addr, data, len);
 		}
-		eprintf ("r_fs_write: file->p->write is null\n");
+		R_LOG_ERROR ("null file->p->write");
 	}
 	return -1;
 }
 
 R_API int r_fs_read(RFS* fs, RFSFile* file, ut64 addr, int len) {
-	if (len < 1) {
-		eprintf ("r_fs_read: too short read\n");
-		return -1;
-	}
-	if (fs && file) {
-		if (file->p && file->p->read) {
-			if (!file->data) {
-				free (file->data);
-				file->data = calloc (1, len + 1);
-			}
-			return file->p->read (file, addr, len);
-		} else {
-			eprintf ("r_fs_read: file->p->read is null\n");
+	R_RETURN_VAL_IF_FAIL (fs && file && len > 0, -1);
+	if (file->p && file->p->read) {
+		if (!file->data) {
+			free (file->data);
+			file->data = calloc (1, len + 1);
 		}
+		return file->p->read (file, addr, len);
 	}
+	R_LOG_ERROR ("null file->p->read");
 	return -1;
 }
 
 R_API RList* r_fs_dir(RFS* fs, const char* p) {
+	R_RETURN_VAL_IF_FAIL (fs && p, NULL);
 	RList *ret = NULL;
 	RFSRoot* root;
 	RListIter* iter;
-	const char* dir;
 	char* path = strdup (p);
 	r_str_trim_path (path);
 	RList *roots = r_fs_root (fs, path);
 	r_list_foreach (roots, iter, root) {
-		if (root) {
-			if (strlen (root->path) == 1) {
-				dir = path;
-			} else {
-				dir = path + strlen (root->path);
-			}
+		if (root && root->p && root->p->dir) {
+			const char *dir = r_str_nlen (root->path, 2) == 1
+				? path: path + strlen (root->path);
 			if (!*dir) {
 				dir = "/";
 			}
@@ -332,24 +343,23 @@ R_API RList* r_fs_dir(RFS* fs, const char* p) {
 			}
 		}
 	}
-	free (roots);
+	r_list_free (roots);
 	free (path);
 	return ret;
 }
 
-R_API int r_fs_dir_dump(RFS* fs, const char* path, const char* name) {
-	RList* list;
+R_API bool r_fs_dir_dump(RFS* fs, const char* path, const char* name) {
+	R_RETURN_VAL_IF_FAIL (fs && path && name, false);
 	RListIter* iter;
-	RFSFile* file, * item;
-	char* str, * npath;
+	RFSFile *file, *item;
 
-	list = r_fs_dir (fs, path);
+	RList *list = r_fs_dir (fs, path);
 	if (!list) {
 		return false;
 	}
 	if (!r_sys_mkdir (name)) {
 		if (r_sys_mkdir_failed ()) {
-			eprintf ("Cannot create \"%s\"\n", name);
+			R_LOG_ERROR ("Cannot create \"%s\"", name);
 			return false;
 		}
 	}
@@ -357,21 +367,9 @@ R_API int r_fs_dir_dump(RFS* fs, const char* path, const char* name) {
 		if (!strcmp (file->name, ".") || !strcmp (file->name, "..")) {
 			continue;
 		}
-		str = (char*) malloc (strlen (name) + strlen (file->name) + 2);
-		if (!str) {
-			return false;
-		}
-		strcpy (str, name);
-		strcat (str, "/");
-		strcat (str, file->name);
-		npath = malloc (strlen (path) + strlen (file->name) + 2);
-		if (!npath) {
-			free (str);
-			return false;
-		}
-		strcpy (npath, path);
-		strcat (npath, "/");
-		strcat (npath, file->name);
+		char *str = r_str_newf ("%s/%s", name, file->name);
+		char *npath = r_str_newf ("%s/%s", path, file->name);
+
 		switch (file->type) {
 		// DON'T FOLLOW MOUNTPOINTS
 		case R_FS_FILE_TYPE_DIRECTORY:
@@ -401,18 +399,15 @@ R_API int r_fs_dir_dump(RFS* fs, const char* path, const char* name) {
 }
 
 static void r_fs_find_off_aux(RFS* fs, const char* name, ut64 offset, RList* list) {
-	RList* dirs;
 	RListIter* iter;
 	RFSFile* item, * file;
-	char* found = NULL;
-
-	dirs = r_fs_dir (fs, name);
+	RList* dirs = r_fs_dir (fs, name);
 	r_list_foreach (dirs, iter, item) {
 		if (!strcmp (item->name, ".") || !strcmp (item->name, "..")) {
 			continue;
 		}
 
-		found = (char*) malloc (strlen (name) + strlen (item->name) + 2);
+		char *found = (char*) malloc (strlen (name) + strlen (item->name) + 2);
 		if (!found) {
 			break;
 		}
@@ -447,12 +442,11 @@ R_API RList* r_fs_find_off(RFS* fs, const char* name, ut64 off) {
 }
 
 static void r_fs_find_name_aux(RFS* fs, const char* name, const char* glob, RList* list) {
-	RList* dirs;
 	RListIter* iter;
 	RFSFile* item;
 	char* found;
 
-	dirs = r_fs_dir (fs, name);
+	RList* dirs = r_fs_dir (fs, name);
 	r_list_foreach (dirs, iter, item) {
 		if (r_str_glob (item->name, glob)) {
 			found = (char*) malloc (strlen (name) + strlen (item->name) + 2);
@@ -482,6 +476,7 @@ static void r_fs_find_name_aux(RFS* fs, const char* name, const char* glob, RLis
 }
 
 R_API RList* r_fs_find_name(RFS* fs, const char* name, const char* glob) {
+	R_RETURN_VAL_IF_FAIL (fs && name && glob, NULL);
 	RList* list = r_list_newf (free);
 	if (list) {
 		r_fs_find_name_aux (fs, name, glob, list);
@@ -490,6 +485,7 @@ R_API RList* r_fs_find_name(RFS* fs, const char* name, const char* glob) {
 }
 
 R_API RFSFile* r_fs_slurp(RFS* fs, const char* path) {
+	R_RETURN_VAL_IF_FAIL (fs && path, NULL);
 	RFSFile* file = NULL;
 	RFSRoot* root;
 	RList* roots = r_fs_root (fs, path);
@@ -502,25 +498,22 @@ R_API RFSFile* r_fs_slurp(RFS* fs, const char* path) {
 			file = root->p->open (root, path, false);
 			if (file) {
 				root->p->read (file, 0, file->size); //file->data
-			}else {
-				eprintf ("r_fs_slurp: cannot open file\n");
+			} else {
+				R_LOG_ERROR ("cannot open file");
 			}
 		} else {
 			if (root->p->slurp) {
 				free (roots);
 				return root->p->slurp (root, path);
 			}
-			eprintf ("r_fs_slurp: null root->p->slurp\n");
+			R_LOG_ERROR ("null root->p->slurp");
 		}
 	}
 	free (roots);
 	return file;
 }
 
-// TODO: move into grubfs
-#include "../../shlr/grub/include/grubfs.h"
-
-#if USE_GRUB
+#if USE_GRUB && WITH_GPL
 static int grub_parhook(void* disk, void* ptr, void* closure) {
 	struct grub_partition* par = ptr;
 	RList* list = (RList*) closure;
@@ -544,19 +537,19 @@ static int fs_parhook(void* disk, void* ptr, void* closure) {
 
 static RFSPartitionType partitions[] = {
 	/* LGPL code */
-	{"dos", &fs_part_dos, fs_parhook},
+	{ "dos", &fs_part_dos, fs_parhook},
 #if USE_GRUB
 	/* WARNING GPL code */
 #if !__EMSCRIPTEN__
 // wtf for some reason is not available on emscripten
-	{"msdos", &grub_msdos_partition_map, grub_parhook},
+	{ "msdos", &grub_msdos_partition_map, grub_parhook},
 #endif
-	{"apple", &grub_apple_partition_map, grub_parhook},
-	{"sun", &grub_sun_partition_map, grub_parhook},
-	{"sunpc", &grub_sun_pc_partition_map, grub_parhook},
-	{"amiga", &grub_amiga_partition_map, grub_parhook},
-	{"bsdlabel", &grub_bsdlabel_partition_map, grub_parhook},
-	{"gpt", &grub_gpt_partition_map, grub_parhook},
+	{ "apple", &grub_apple_partition_map, grub_parhook},
+	{ "sun", &grub_sun_partition_map, grub_parhook},
+	{ "sunpc", &grub_sun_pc_partition_map, grub_parhook},
+	{ "amiga", &grub_amiga_partition_map, grub_parhook},
+	{ "bsdlabel", &grub_bsdlabel_partition_map, grub_parhook},
+	{ "gpt", &grub_gpt_partition_map, grub_parhook},
 #endif
 	// XXX: In BURG all bsd partition map are in bsdlabel
 	//{ "openbsdlabel", &grub_openbsd_partition_map },
@@ -572,11 +565,8 @@ R_API const char* r_fs_partition_type_get(int n) {
 	return partitions[n].name;
 }
 
-R_API int r_fs_partition_get_size(void) {
-	return R_FS_PARTITIONS_LENGTH;
-}
-
 R_API RList* r_fs_partitions(RFS* fs, const char* ptype, ut64 delta) {
+	R_RETURN_VAL_IF_FAIL (fs && ptype, NULL);
 	int i, cur = -1;
 	for (i = 0; partitions[i].name; i++) {
 		if (!strcmp (ptype, partitions[i].name)) {
@@ -602,27 +592,34 @@ R_API RList* r_fs_partitions(RFS* fs, const char* ptype, ut64 delta) {
 		{
 #endif
 			RFSPartitionIterator iterate = partitions[i].ptr;
-			iterate (fs, partitions[i].iterate, list); //grub_parhook, list);
+			iterate (fs, partitions[i].iterate, list);
 		}
 		return list;
 	}
-	if (ptype && *ptype) {
-		eprintf ("Unknown partition type '%s'.\n", ptype);
+	if (R_STR_ISNOTEMPTY (ptype)) {
+		R_LOG_ERROR ("Unknown partition type '%s'. Use 'mL' command to list them all", ptype);
 	}
-	eprintf ("Supported types:\n");
-	for (i = 0; partitions[i].name; i++) {
-		eprintf (" %s", partitions[i].name);
-	}
-	eprintf ("\n");
 	return NULL;
 }
 
 R_API int r_fs_partition_type_str(const char* type) {
-	// TODO: implement
+#if USE_GRUB && WITH_GPL
+	// TODO: properly implement our types to not depend on grub
+	if (!strcmp (type, "fat")) {
+		return GRUB_PC_PARTITION_TYPE_FAT32;
+	}
+	if (!strcmp (type, "ext2")) {
+		return GRUB_PC_PARTITION_TYPE_EXT2FS;
+	}
+	if (!strcmp (type, "hfs")) {
+		return GRUB_PC_PARTITION_TYPE_HFS;
+	}
+#endif
 	return 0;
 }
 
 R_API const char* r_fs_partition_type(const char* part, int type) {
+#if USE_GRUB && WITH_GPL
 	// XXX: part is ignored O_o
 	switch (type) {
 	case GRUB_PC_PARTITION_TYPE_FAT12:
@@ -632,24 +629,18 @@ R_API const char* r_fs_partition_type(const char* part, int type) {
 	case GRUB_PC_PARTITION_TYPE_FAT32_LBA:
 	case GRUB_PC_PARTITION_TYPE_FAT16_LBA:
 		return strdup ("fat");
-
 	case GRUB_PC_PARTITION_TYPE_EXT2FS:
 		return strdup ("ext2");
-
 	case GRUB_PC_PARTITION_TYPE_MINIX:
 	case GRUB_PC_PARTITION_TYPE_LINUX_MINIX:
 		return strdup ("minix");
-
 	case GRUB_PC_PARTITION_TYPE_NTFS:
 		return strdup ("ntfs");
-
 	case GRUB_PC_PARTITION_TYPE_EXTENDED:
 	case GRUB_PC_PARTITION_TYPE_LINUX_EXTENDED:
-		return strdup ("ext3");
-
+		return strdup ("ext3"); // XXX
 	case GRUB_PC_PARTITION_TYPE_HFS:
 		return strdup ("hfs");
-
 	case GRUB_PC_PARTITION_TYPE_WIN95_EXTENDED: // fat?
 	case GRUB_PC_PARTITION_TYPE_EZD:
 	case GRUB_PC_PARTITION_TYPE_VSTAFS:
@@ -662,14 +653,17 @@ R_API const char* r_fs_partition_type(const char* part, int type) {
 	default:
 		return NULL;
 	}
+#endif
+	return NULL;
 }
 
 R_API char* r_fs_name(RFS* fs, ut64 offset) {
+	R_RETURN_VAL_IF_FAIL (fs, NULL);
 	ut8 buf[1024];
 	int i, j, len, ret = false;
 
 	for (i = 0; fstypes[i].name; i++) {
-		RFSType* f = &fstypes[i];
+		const RFSType* f = &fstypes[i];
 		len = R_MIN (f->buflen, sizeof (buf) - 1);
 		fs->iob.read_at (fs->iob.io, offset + f->bufoff, buf, len);
 		if (f->buflen > 0 && !memcmp (buf, f->buf, f->buflen)) {
@@ -692,23 +686,24 @@ R_API char* r_fs_name(RFS* fs, ut64 offset) {
 }
 
 R_API void r_fs_view(RFS* fs, int view) {
+	R_RETURN_IF_FAIL (fs);
 	fs->view = view;
 }
 
 R_API bool r_fs_check(RFS *fs, const char *p) {
+	R_RETURN_VAL_IF_FAIL (fs && p, false);
 	RFSRoot *root;
 	RListIter *iter;
-	char* path = strdup (p);
-	if (!path) {
-		return false;
-	}
-	r_str_trim_path (path);
-	r_list_foreach (fs->roots, iter, root) {
-		if (r_fs_match (path, root->path, strlen (root->path))) {
-			free (path);
-			return true;
+	char *path = strdup (p);
+	if (path) {
+		r_str_trim_path (path);
+		r_list_foreach (fs->roots, iter, root) {
+			if (r_fs_match (path, root->path, strlen (root->path))) {
+				free (path);
+				return true;
+			}
 		}
+		free (path);
 	}
-	free (path);
 	return false;
 }
